#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/sendfile.h>
#include <dirent.h>
#include <ctype.h>

#define MAXLEN 4096

static char *mime_types[] = {
    /* Mime des elements web les plus communs*/
    "htm",      "text/html",
    "html",     "text/html",
    "shtml",    "text/html",
    "css",      "text/css",
    "js",       "application/x-javascript",
    "txt",      "text/plain",
    "c",        "text/plain",
    "h",        "text/plain",
 
    /*Mime des images*/
    "ico",  "image/x-icon",
    "bmp",  "image/bmp",
    "gif",  "image/gif",
    "jpe",  "image/jpeg",
    "jpeg", "image/jpeg",
    "jpg",  "image/jpeg",
    "png",  "image/png",
    "tif",  "image/tiff",
    "tiff", "image/tiff",
 
    /* Mime des fichiers d'applications usuels*/
    "tex",  "application/x-tex",
    "ps",   "application/postscript",
    "pdf",  "application/pdf",
    "zip",  "application/zip",
    "tgz",  "application/x-compressed",
    "gz",   "application/x-gzip",
    "rtf",  "application/rtf",
    "sh",   "application/x-sh",
    "tar",  "application/x-tar",
    "z",    "application/x-compress",
    "ogg",  "application/ogg",
    "pac",  "application/x-ns-proxy-autoconfig",
 
    /* Mime des fichiers audios usuels*/
    "mp3",  "audio/mpeg",
    "m3u",  "audio/x-mpegurl",
    "ra",   "audio/x-pn-realaudio",
    "wav",  "audio/x-wav",
    "au",   "audio/basic",
    "mid",  "audio/midi",
    "midi", "audio/midi",
 
    /* Mime des fichiers videos usuels*/
    "asf",  "video/x-ms-asf",
    "avi",  "video/x-msvideo",
    "mov",  "video/quicktime",
    "mp2",  "video/mpeg",
    "mpe",  "video/mpeg",
    "mpeg", "video/mpeg",
    "mpg",  "video/mpeg",
    "qt",   "video/quicktime",

    /* Mime zabri */
	"vrml", "model/vrml"
};
 
static const char * get_mime_type(const char *filename){
  int i, size;
  char* extension;
  
  extension = strrchr(filename, '.');
  if (!extension) return "text/plain";
  extension++;

  size = sizeof(mime_types)/sizeof(*mime_types);
  for( i = 0; i < size; i += 2){
    if( !strcmp(extension, mime_types[i]))
      return mime_types[i+1];
  }
  return "application/octet-stream";
}

static int hexit( char c )
{
    if ( c >= '0' && c <= '9' )
	return c - '0';
    if ( c >= 'a' && c <= 'f' )
	return c - 'a' + 10;
    if ( c >= 'A' && c <= 'F' )
	return c - 'A' + 10;
    return 0;		/* shouldn't happen, we're guarded by isxdigit() */
}

static void strdecode( char* to, char* from )
{
    for ( ; *from != '\0'; ++to, ++from )
	{
	if ( from[0] == '%' && isxdigit( from[1] ) && isxdigit( from[2] ) )
	    {
	    *to = hexit( from[1] ) * 16 + hexit( from[2] );
	    from += 2;
	    }
	else
	    *to = *from;
	}
    *to = '\0';
}

int is_dir (const char* path)
{
    struct stat st;

    if(lstat(path, &st) == -1)
        return 0;
    return S_ISDIR(st.st_mode);
    /* return ((st.st_mode & S_IFMT) == SIFLINK); */
}

static void write_http_error (int sock, int code, const char* phrase)
{
    char buf[MAXLEN];
    int null;
    null = snprintf(buf, MAXLEN, "%s %d %s\r\n", "HTTP/1.0", code, phrase);
    null = write(sock,buf,strlen(buf));
}

static void write_http_headers (int sock, const char* ctype, const char* clen)
{
    char buf[MAXLEN];
    int null;
    null = snprintf(buf, MAXLEN, "%s\r\n%s\r\n\r\n", ctype, clen);
    null = write(sock, buf, strlen(buf));
}

static void content_type(char* buf, const char* file) {
    snprintf(buf, MAXLEN, "Content-Type: %s", get_mime_type(file));
}

static void content_length(char* buf, int flen) {
    snprintf(buf, MAXLEN, "Content-Length: %d", flen);
}

static void dir_output(int sock, const char* path, const char* filename)
{
    char buffer[MAXLEN];
    char pathname[MAXLEN];
    char* p;
    int null;

    strncpy(pathname, path, MAXLEN);
    pathname[MAXLEN-1]='\0'; /* cut */
    p = &(pathname[strlen(pathname)]);
    strncpy(p, filename, MAXLEN - 1 - strlen(pathname));
    p[MAXLEN-1-strlen(pathname)]='\0'; /* we never know */
    
    if (is_dir(pathname))
        snprintf(buffer, MAXLEN, "<a href=\"%s/\">%s/</a><br>\n", filename, filename);
    else
        snprintf(buffer, MAXLEN, "<a href=\"%s\">%s</a><br>\n", filename, filename);
    null = write(sock, buffer, strlen(buffer));
}

static void serve_directory_list( int sock, const char* path)
{
        struct dirent **namelist;
        int n, null, i=0;
        char buf[MAXLEN];

        buf[0]='.';
        strncpy(buf+1, path, MAXLEN-1);
        buf[MAXLEN-2]='\0'; /* we never know */
        n = scandir(buf, &namelist, 0, alphasort);
        if (n < 0) {
            write_http_error(sock, 500, "Internal Error\r\n");
        }
        else {
            write_http_error(sock, 200, "Ok");
            write_http_headers(sock, "Content-Type: text/html", "Connection: close");
            null = write(sock,"<html><title>directory</title><body>", 36);
            while(i < n) {
                dir_output(sock, buf, namelist[i]->d_name);
                free (namelist [i]);
                i++;
            }
            free (namelist);
            null = write(sock,"</body></html>", 14);
        }
}

int launch_web_server(int sock) {
    /* read GET /path HTTP/1.0... through sock */
    char *buffer = NULL;
    char *p, *method, *path;
    char *file, ctype[MAXLEN], clen[MAXLEN];
    struct stat fstats;
    int ret = 0;
    int fd, i;
    
    buffer = malloc(1); buffer[0]='\0';
    i=1;
    do {
        buffer=realloc(buffer, strlen(buffer)+MAXLEN+1);
        ret += read(sock, buffer + ret, MAXLEN);
        buffer[ret]='\0'; /* force to 0 for strstr() */
        i++;
    } while (i < MAXLEN && (!strstr(buffer, "\r\n\r\n")));

    p = strchr(buffer, ' ');
    p[0] = '\0';
    method = malloc(8);
    method = strncpy(method, buffer, 8);
    method[7]='\0'; /* we never know */
    path = malloc(MAXLEN);
    path = strncpy(path, p+1, MAXLEN);
    path[MAXLEN-1]='\0'; /* force to 0 for strchr() and strstr() */
    p = strchr(path, ' ');
    p[0] = '\0';
    free(buffer);
    
    if (path[0] != '/') {
        write_http_error(sock, 400, "Bad Request\r\n");
        close(sock);
        _exit(EXIT_FAILURE);
    }
    if (strstr(path, "..")||strstr(path,"%2E%2E")) { /* attempt to go up ? */
        write_http_error(sock, 403, "Forbidden\r\n");
        close(sock);
        _exit(EXIT_FAILURE);
    }
    if ( !strstr(method, "GET") ) {
        write_http_error(sock, 501, "Not Implemented\r\n");
        close(sock);
        _exit(EXIT_FAILURE);
    }
    /* decode url-encoded string */
    strdecode(path, path);
    /* if resource is a directory, serve file listing */
    if(path[strlen(path)-1] == '/' || is_dir(&path[1])) {
        serve_directory_list(sock, path);
    } else { /* file or pipe or device */
        file = &(path[1]); /* get relative */
        if ( stat( file, &fstats ) < 0 ) {
            write_http_error( sock, 404, "Not Found\r\n");
            close(sock);
            _exit(EXIT_FAILURE);
        }
        fd = open(file, O_RDONLY);
        if (fd < 0) {
            write_http_error(sock, 403, "Forbidden\r\n");
            _exit(EXIT_FAILURE);
        }
        write_http_error(sock, 200, "Ok");
        content_type(ctype, file);
        content_length(clen, fstats.st_size);
        write_http_headers(sock, ctype, clen);
        sendfile(sock, fd, 0, fstats.st_size);
        close(fd);
    }
    close(sock);
    free(method);
    free(path);
    _exit(EXIT_SUCCESS);
}

